
# Copyright (C) 2015 Centro de Computacao Cientifica e Software Livre
# Departamento de Informatica - Universidade Federal do Parana
#
# This file is part of portalmec.
#
# portalmec is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# portalmec is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with portalmec.  If not, see <http://www.gnu.org/licenses/>.

# == Schema Information
#
# Table name: users
#
#  id                     :integer          not null, primary key
#  email                  :string           default(""), not null
#  encrypted_password     :string           default(""), not null
#  reset_password_token   :string
#  reset_password_sent_at :datetime
#  remember_created_at    :datetime
#  sign_in_count          :integer          default("0"), not null
#  current_sign_in_at     :datetime
#  last_sign_in_at        :datetime
#  current_sign_in_ip     :string
#  last_sign_in_ip        :string
#  created_at             :datetime         not null
#  updated_at             :datetime         not null
#  provider               :string           default("email"), not null
#  uid                    :string           default(""), not null
#  confirmation_token     :string
#  confirmed_at           :datetime
#  confirmation_sent_at   :datetime
#  unconfirmed_email      :string
#  tokens                 :text
#  name                   :string
#  avatar_file_name       :string
#  avatar_content_type    :string
#  avatar_file_size       :integer
#  avatar_updated_at      :datetime
#  bookmarks_count        :integer          default("0")
#  user_category_id       :integer
#  score                  :float            default("0.0")
#  follows_count          :integer          default("0")
#  description            :text
#

class User < ApplicationRecord
  include Followable
  include Reputationable
  include Subjectable
  include Tagger
  include Complainable
  include Publisher
  include Trackable

  # Include default devise modules. Others available are:
  # :confirmable, :lockable, :timeoutable and :omniauthable
  devise :database_authenticatable, :registerable,
         :recoverable, :rememberable, :trackable, :validatable,
         :omniauthable

  include DeviseTokenAuth::Concerns::User

  belongs_to :school

  has_and_belongs_to_many :roles
  has_and_belongs_to_many :institutions

  has_many :bookmarks, as: :bookmarkable
  has_many :bookmark_collections, through: :bookmarks, source: :bookmarkable, source_type: 'Collection'
  has_many :bookmark_learning_objects, through: :bookmarks, source: :bookmarkable, source_type: 'LearningObject'
  has_many :bookmarks
  #has_many :collections, as: :owner
  #has_many :learning_objects, as: :publisher
  has_many :views
  has_many :downloads
  has_many :likes
  has_many :shares
  has_many :follows
  has_many :reviews
  has_many :applications
  has_many :searches

  has_many :submissions, class_name: "Submission", foreign_key: "submitter_id"
  has_many :curations, class_name: "Submission", foreign_key: "curator_id"
  has_many :curator_assignments
  has_many :assignments, through: :curator_assignments, source: :submission

  after_create :default_role
  before_save :verify_teacher_id
  after_save :verify_dspace_info
  before_destroy :delete_index

  has_attached_file :avatar, styles: { medium: '300x300>', thumb: '60x60>' }, default_url: ''
  validates_attachment_content_type :avatar, content_type: ['image/jpg', 'image/jpeg', 'image/png', 'image/gif']

  has_attached_file :cover, styles: { medium: '800x300>', thumb: '160x60>' }, default_url: ''
  validates_attachment_content_type :cover, content_type: ['image/jpg', 'image/jpeg', 'image/png', 'image/gif']

  validates :terms_of_service, acceptance: true

  validates :dspace_url, presence: true, if: Proc.new {|u| !u.dspace_handle.blank? || !u.dspace_sets.blank?}

  validates :dspace_handle, presence: true, if: Proc.new {|u| !u.dspace_url.blank? || !u.dspace_sets.blank?}

  enum submitter_request: { default: 0, requested: 1, accepted: 2, rejected: 3 }

  validates :school_id, presence: true, unless: Proc.new {|u| u.default? || u.rejected?}

  enum state: { active: 0, blocked: 1, banished: 2}

  searchkick language: 'brazilian', match: :word_start, searchable: [:name], callbacks: :async, index_prefix: Socket.gethostname.downcase

  acts_as_paranoid

  # don't versionate frequently changed fields
  has_paper_trail ignore: [:tokens, :sign_in_count, :current_sign_in_at, :last_sign_in_at]

  def search_data
    {
      name: name,
      score: score
    }
  end

  def name=(name)
    write_attribute(:name, name.strip) if !name.blank?
  end

  def should_index?
      deleted_at.nil?
  end

  def delete_index
    User.searchkick_index.remove(self)
  end

  def is_admin?
    roles.each do |role|
      return true if role.name == 'admin'
    end
    false
  end

  def is_moderator?
    roles.each do |role|
      return true if role.name == 'moderator'
    end
    false
  end

  def is_curator?
    roles.each do |role|
      return true if role.name == 'curator'
    end
    false
  end

  def is_supervisor?
    roles.each do |role|
      return true if role.name == 'supervisor'
    end
    false
  end

  def is_editor?
    roles.each do |role|
      return true if role.name == 'editor'
    end
    false
  end

  def is_submitter?
    roles.each do |role|
      return true if role.name == 'submitter'
    end
    false
  end

  def is_publisher?
    roles.each do |role|
      return true if role.name == 'publisher'
    end
    false
  end

  def is_partner?
    roles.each do |role|
      return true if role.name == 'partner'
    end
    false
  end

  def can_edit?
    self.is_admin? || self.is_editor?
  end

  def active_for_authentication?
    super && self.active?
  end

  def associated_collections
    c = collections.blank? ? [] : collections.to_a
    institutions.each { |i| c.push(*i.collections.to_a) }
    c
  end

  def associated_learning_objects
    lo = learning_objects.blank? ? [] : learning_objects.to_a
    institutions.each { |i| lo.push(*i.learning_objects.to_a) }
    lo
  end

  def own_reviews
    Review.where(user_id: id)
  end

  def token_validation_response
    UserDeviseSerializer.new(self, root: false)
  end


  def received_reviews
    Review.joins('INNER JOIN learning_objects ON ' \
                  "learning_objects.publisher_id = #{id} AND " \
                  'reviews.reviewable_id = learning_objects.id')
  end

  # ~~~~ followable actions ~~~~ #
  # An user can follow anothers users and collections
  # Examples:
  # current_user.follow(User.find_by_email('john@revelator.com'))
  # current_user.follow(Collection.find(143))
  def follow(followable)
    Follow.create(user: self, followable: followable)
  end

  # Examples:
  # current_user.unfollow(User.find_by_email('john@revelator.com'))
  # current_user.unfollow(Collection.find(143))
  def unfollow(followable)
    Follow.where(user: self, followable: followable).destroy_all
  end

  # An user can see if is following another user
  # Examples:
  # current_user.following?(User.find_by_email('john@revelator.com')) => true
  # current_user.following?(Collection.find(143)) => true
  def following?(followable)
    !follows.where(followable_id: followable.id, followable_type: followable.class.name).blank?
  end

  # This function permits see what person is following.
  # For current user, list all users and all collections. But,
  # if, isn't current user, list only users, and public collections which this person follows.
  def following(followable_type, is_current_user = true)
    f = Follow.where(user_id: id, followable_type: followable_type).all

    # If type collection, and isn't current_user, so should list only public collections
    if followable_type.to_s == 'Collection' && !is_current_user
      f = Follow.joins('INNER JOIN collections ON follows.followable_id = collections.id').where(
        "follows.user_id = #{id} " \
        'AND follows.followable_type = \'Collection\' ' \
        'AND collections.id = follows.followable_id ' \
        'AND collections.privacy = \'public\''
      )
    end

    f
  end

  # Returns everything that this user have followed or created
  def watching
    watching_array = []
    [User, Collection].each do |model|
      following(model.to_s).each do |follow|
        watching_array.push(*model.find(follow.followable_id))
      end
    end
    watching_array.push(*collections)
    watching_array.push(*learning_objects)
    watching_array.push(*reviews)
    watching_array.push(*submissions)
    watching_array.push(*curator_assignments)
  end

  # ~~~~ end followable actions ~~~~ #

  def default_role
    roles << Role.submitter
  end

  def self.from_omniauth(access_token)
    user = User.find_by(email: user_email)
    return nil if user.blank?

    data = access_token.info
    user_email = access_token.provider == 'twitter' ? "#{access_token.uid}@twitter.com" : data['email']
    User.create(
      name: data['name'],
      email: user_email,
      provider: access_token.provider,
      uid: access_token.uid,
      password: Devise.friendly_token[0, 20]
    )
  end

  def verify_dspace_info
    if roles.include?(Role.find_by_name("partner")) && !(changed & ["dspace_url", "dspace_handle", "dspace_sets"]).empty?
      DspaceMailer.dspace_info_updated(self).deliver_now
    end
  end

  def verify_teacher_id
    return true if (changed & ["teacher_id", "birthday"]).empty?

    params = {
      filter: "teacher:#{teacher_id},day: #{birthday.day},month:#{birthday.month},year:#{birthday.year},min_year:2017"
    }
    uri = URI("https://simcaq.c3sl.ufpr.br/api/v1/verify_teacher")
    uri.query = URI.encode_www_form(params)
    response = JSON.parse(Net::HTTP.get(uri))
    if !response['result'].blank?
      roles << Role.publisher
      return true
    end

    errors.add(:INEP, "Dados do INEP incorretos")
    return false
  end

  def activity_owner
    self
  end

  def ignore_changes
    super + %w(tokens sign_in_count current_sign_in_at last_sign_in_at current_sign_in_ip last_sign_in_ip score follows_count confirmation_token confirmed_at confirmation_sent_at)
  end

  def update_tracked_fields(request)
    super
    # Change request.remote_ip to request.env["HTTP_X_REAL_IP"] in production
    self.current_sign_in_ip = request.remote_ip
  end

  def treat_complaintment
  end

  def complaint_accept(option)
    if option == "blocked"
      self.suspended_at = Time.current
      self.reactivated_at = self.suspended_at + 7.day
      self.increment(:times_blocked, 1)
      self.blocked!
    elsif option == "banished"
      self.suspended_at = Time.current
      self.reactivated_at = nil
      self.increment(:times_blocked, 1)
      self.banished!
    elsif
      raise ArgumentError.new(I18n.t("error.wrong_arguments", param: "option"))
    end
  end

  def complaint_reject(params)
  end
end
